from abc import ABC, abstractmethod
from typing import TYPE_CHECKING, Any, Generic, Tuple, Type, TypeVar, Union

from .bitstring import BitString
from .typing import RegisterState

if TYPE_CHECKING:
    from .blocks import RegisterBlock


ParserType = TypeVar("ParserType")


class RegisterParser(ABC, Generic[ParserType]):
    """
    Abstract base class for data descriptors that parse/set register values.
    Subclasses represent the various formats of registers used by a Device.
    """

    def __init__(self, address: Union[int, slice]):
        """
        Initialize a parser instance for the given register address(es).

        Parameters
        ----------
        address : Union[int, slice]
            An index or slice to retrieve byte(s) at a given address
        """
        if isinstance(address, int):
            address = slice(address, address + 1)
        self._address = address
        # FIXME if there are multiple instances of a given RegisterBlock
        #  class, all their parsers will access the same state! this could
        #  cause weird behavior; store state in a dict keyed by instance?
        #  or use a meta-class to define a label?
        #  https://nbviewer.jupyter.org/urls/gist.github.com/ChrisBeaumont/5758381/raw/descriptor_writeup.ipynb#Make-sure-to-keep-instance-level-data-instance-specific
        self._register_state: RegisterState = []

    def __get__(
        self,
        instance: "RegisterBlock[Any]",
        owner: Type["RegisterBlock[Any]"],
    ) -> ParserType:
        """
        When accessing a RegisterParser, set its register state based on the
        state of the RegisterBlock instance that it belongs to (which in turn
        reads from the I2CDevice instance /it/ belongs to). See note in
        RegisterBlock.__get__ about why this behavior is needed.
        """
        self._register_state = instance.register_state
        return self._value()

    def __set__(
        self, instance: "RegisterBlock[Any]", value: ParserType
    ) -> None:
        """
        When setting a RegisterParser value, make sure its state is in sync
        with all pending changes to the RegisterBlock instance that it belongs
        to, then update the pending state with changes from this parser.
        """
        self._register_state = instance.pending_state
        new_state = self._prepare_update(value)
        instance.update_register_state(self._address, new_state)

    @abstractmethod
    def _prepare_update(self, value: ParserType) -> RegisterState:
        """
        Subclasses should implement a _prepare_update method which returns a
        register address offset (index or slice) and a list of BitStrings to
        replace the data currently at the address(es).
        """

    @abstractmethod
    def _value(self) -> ParserType:
        """
        Subclasses should implement a _value method which returns an object
        of the appropriate type to represent the register content
        """

    @property
    def _local_bytes(self) -> RegisterState:
        """
        Returns a slice of the register state with just those bytes needed for
        an instance to do its parsing.
        """
        return self._register_state[self._address]


class BCDParser(RegisterParser[int]):
    """
    Base class for registers with simple BCD format. Subclasses can define a
    class attribute `bcd_bounds` with a tuple of (start, stop) tuples (i.e. a
    tuple representing the bits to slice for each digit). By default, parses
    2 decimal digits from the 4-bit nibbles of a byte.
    """

    bcd_bounds: Tuple[Tuple[int, int], ...] = ((0, 4), (4, 8))

    def _value(self) -> int:
        return self._local_bytes[0].decode_bcd(*self.bcd_bounds)

    def _prepare_update(self, value: int) -> RegisterState:
        offset = min(s[0] for s in self.bcd_bounds)
        sizes = [s[1] - s[0] for s in self.bcd_bounds]
        new_bitstring = BitString.encode_bcd(value, sizes)
        new_state = [self._local_bytes[0].replace(offset, new_bitstring)]
        return new_state


class FlagParser(RegisterParser[bool]):
    """
    Base class for flags stored as a single bit in a register.
    """

    def __init__(self, address: int, bit: int) -> None:
        # FIXME better as extra arg to constructor or class-level attribute?
        #  for BCDParser, its nice to have subclasses define their bounds
        #  in a way that can be reused across multiple registers. i'm not sure
        #  there is an equivalent use case for FlagParser, but there could
        #  be. it feels weird for them to implement very similar functionality
        #  in 2 different ways.
        self._bit = bit
        super().__init__(address)

    def _value(self) -> bool:
        return bool(self._local_bytes[0][self._bit])

    def _prepare_update(self, value: bool) -> RegisterState:
        new_state = [
            self._local_bytes[0].replace(self._bit, BitString(int(value)))
        ]
        return new_state



from abc import ABC, abstractmethod
from datetime import datetime
from typing import Generic, Type, TypeVar, Union

from .devices import I2CDevice
from .parsers import RegisterParser
from .typing import RegisterState

BlockType = TypeVar("BlockType")


class RegisterBlock(Generic[BlockType], ABC):
    """
    Abstract base class for collections of registers that represent distinct
    features of an I2C device. A RegisterBlock translates between high-level
    data structures and the low-level representation of that data as expressed
    by RegisterParsers. For example, for the DS series RTCs, there are sub-
    classes of RegisterBlock for the clock, the alarms, and their configuration
    states. The Clock subclass encapsulates RegisterParsers for the BCD-ish
    encoding of the Hour, Minute, Second, etc. stored in the device registers.

    RegisterBlock is a Generic type. When subclassing, add the appropriate type
    for the value represented by the subclass to its signature:

      class TimekeepingRegisterBlock(RegisterBlock[datetime]): ...

    A RegisterBlock subclass should define one or more attributes that are
    RegisterParsers. Subclasses must also define two methods:

     1) `_value` to read the data from its attributes and produce a value of
        the designated type
     2) `_prepare_update` to set its attributes to a given value

    For example, suppose some device stored a positive decimal number like
    12.34 with the integer part in register 0x00 and the fractional part in
    register 0x01, each represented as 2 digit standard BCD. You want to read
    or write this value as a 2-tuple of ints. A RegisterBlock for accessing
    this number could be:

        class DecimalRegisterBlock(RegisterBlock[Tuple[int, int]]):
            integer_part = BCDRegisterParser(0x00)
            fractional_part = BCDRegisterParser(0x01)

            def _value(self) -> Tuple[int, int]:
                return self.integer_part, self.fractional_part

            def _prepare_update(self, value: Tuple[int, int]) -> None:
                self.integer_part, self.fractional_part = value
    """

    @property
    def register_state(self) -> "RegisterState":
        """
        Accesses register state from the most recent read of the parent device.
        """
        return self._register_state

    @register_state.setter
    def register_state(self, state: "RegisterState") -> None:
        """
        Setting register_state also keeps a copy to use as pending_state.
        """
        self._register_state = state
        self.pending_state = self._register_state.copy()

    def __init__(self) -> None:
        """
        Initialize a new RegisterBlock. RegisterBlock is a data descriptor, so
        it must be used as an attribute on a subclass of I2CDevice in order to
        have access to the device register state.
        """
        # The very first access to the descriptor will populate actual state.
        self.register_state: RegisterState = []

    def __get__(
        self, instance: "I2CDevice", owner: Type["I2CDevice"]
    ) -> BlockType:
        """
        RegisterBlock is a data descriptor with access to the state of the
        I2CDevice instance that it belongs to, so we can use that register
        state for all parsers associated with this RegisterBlock (see
        RegisterParser.__get__).

        It is important for all RegisterParser instances to have a shared
        register state (i.e. the state stored in this class) in order to avoid
        mistakes if the state changes during a read. For example, if an RTC's
        Second register is read at 0 minutes 59 seconds, and then the clock
        ticks before we read the Minute register, the time would come out as
        1 minute 59 seconds. Maxim DS RTCs (and probably others) use of 2 sets
        of registers to prevent this issue from affecting I2C block reads, so
        we just need to make sure we only make one call to `read_registers()`
        for all the RegisterParsers within a RegisterBlock.
        """
        if not instance:
            raise AttributeError(
                "RegisterBlock must be accessed from an I2CDevice instance."
            )
        self.register_state = instance.read_registers()
        return self._value()

    def __set__(self, instance: "I2CDevice", value: BlockType) -> None:
        """
        Setting the value of the RegisterBlock updates its state via the
        RegisterParser descriptors that belong to the block.
        """
        # Make sure we have the latest state loaded before modifying it
        self.register_state = instance.read_registers()
        self._prepare_update(value)
        # A minor optimization to only write a contiguous block from the first
        # changed register to the last changed register, leaving the rest
        # unmodified. This helps improve the speed of small updates.
        addresses_changed = [
            i
            for i, b in enumerate(self.pending_state)
            if b != self._register_state[i]
        ]
        first_changed = min(addresses_changed)
        last_changed = max(addresses_changed)
        to_write = self.pending_state[first_changed : last_changed + 1]
        instance.write_registers(to_write, first_changed)

    @abstractmethod
    def _prepare_update(self, value: BlockType) -> None:
        """
        Subclasses should define behavior for setting the values of their
        RegisterParser attributes to reflect the requested `value` for the
        RegisterBlock. Parsers' `__set__` methods call `update_register_state`
        on this instance so they can all keep their pending state in sync.
        """

    @abstractmethod
    def _value(self) -> BlockType:
        """
        Value should return an appropriate object to represent the state of
        this register block e.g. a datetime for the clock/alarms or a float for
        the temperature
        """

    def update_register_state(
        self, address: Union[int, slice], value: "RegisterState"
    ) -> None:
        """
        RegisterParsers should call this method to stage their changes to the
        register state. This allows parsers to be aware of each other's pending
        changes so e.g. two distinct parsers can flip two different bits in the
        same register. Once all parsers have staged their changes (implement
        via _prepare_update), the __set__ method will write all the changes to
        the parent I2CDevice instance.

        Parameters
        ----------
        address : Union[int, slice]
            The register address(es) to set

        value : RegisterState
            The bytes to insert at address
        """
        if isinstance(address, int):
            address = slice(address, address + 1)
        if len(value) != len(self.pending_state[address]):
            raise ValueError("Value must have as many bytes as slice")
        self.pending_state[address] = value


class DatetimeRegisterBlock(RegisterBlock[datetime]):
    """
    Base class whose subclasses keep track of the register addresses where
    various components of the date/time/alarms are stored for RTC ICs such
    as the Maxim DS series.
    """

    hour: RegisterParser[int]
    minute: RegisterParser[int]
    day_of_month: RegisterParser[int]

    # Define defaults for attributes that may be left unset, e.g. the DS3231
    # and DS1337 have no seconds for Alarm 2, and no year or month for either
    # Alarm.
    @property
    def second(self) -> Union[RegisterParser[int], int]:
        return 0

    @second.setter
    def second(self, value: int) -> None:
        pass

    @property
    def month(self) -> Union[RegisterParser[int], int]:
        return datetime.now().month

    @month.setter
    def month(self, value: int) -> None:
        pass

    @property
    def year(self) -> Union[RegisterParser[int], int]:
        return datetime.now().year

    @year.setter
    def year(self, value: int) -> None:
        pass

    def _prepare_update(self, value: datetime) -> None:
        # FIXME pycharm doesn't understand you can assign an int to the
        #  parser descriptors, but mypy does
        self.second = value.second
        self.minute = value.minute
        self.hour = value.hour
        self.day_of_month = value.day
        self.month = value.month
        self.year = value.year

    def _value(self) -> datetime:
        try:
            value = datetime(
                self.year,
                self.month,
                self.day_of_month,
                self.hour,
                self.minute,
                self.second,
            )
        except ValueError as err:
            raise ValueError(
                "Could not parse datetime. Perhaps the register state is"
                "invalid? Try setting to a known valid state first."
            ) from err
        return value



"""
BitString provides an interface to interact with unsigned integers
through their bit-string representations, allowing you to slice, append,
replace, etc. as well as all the usual math you can do with an int.
"""
from typing import Collection, Tuple, Union


class BitString(int):
    """
    Given a binary string or positive integer, store the value as an integer
    which can be sliced or concatenated based on the string representation.

    Useful for working with binary-coded decimal values where, for example,
    each 4-bit nibble of a byte is used to represent a decimal digit - see
    `BitString.encode_bcd` and `BitString.decode_bcd`.

    Examples
    --------
    >>> BitString(10)
    BitString('1010')
    >>> BitString(10)[:2]
    BitString('10')
    >>> BitString(10, fill=8)
    BitString('00001010')
    >>> assert BitString('1010') + BitString('1010') == 10 + 10
    >>> assert BitString('1').append(BitString('1')) == BitString('11')
    """

    bitstring: str  # signal to mypy that this will be added by __new__

    def __new__(
        cls,
        value: Union[str, int],
        fill: int = 0,
        raise_on_overflow: bool = False,
    ) -> "BitString":
        """
        Accepts a positive integer or string representation of a binary
        unsigned integer. Stores the binary string representation.

        Parameters
        ----------
        value : Union[str, int]
            A positive integer or its binary string representation
        fill : int, default 0
            The minimum length of the bitstring (will pad with 0s on the left).
            Note that if you pass in a string, BitString will always maintain
            at least the length of the string you input.

        raise_on_overflow : bool, default False
            Whether to raise ValueError if fill is non-zero and the value
            cannot be represented by the number of bits specified by fill
        """
        if isinstance(value, str):
            # make the fill value at least as long as the input
            fill = max(len(value), fill)
            value = int(value.replace(" ", ""), 2)
        if value < 0:
            raise ValueError("BitString does not support negative values")
        instance = super().__new__(cls, value)
        bitstring = bin(instance)[2:]
        instance.bitstring = bitstring.zfill(fill)
        if raise_on_overflow and len(instance.bitstring) > fill:
            raise ValueError(f"Value {value} too large for {fill} bits")
        return instance

    def __getitem__(self, key: Union[int, slice]) -> "BitString":
        """
        Slice the string representation

        >>> BitString('1010')[:2]
        BitString('10')
        """
        slice_ = self.bitstring[key]
        return BitString(slice_, fill=len(slice_))

    def replace(self, offset: int, value: "BitString") -> "BitString":
        """
        Return a new BitString where `len(value)` bits starting at `offset`
        are replaced by `value`.

        Examples
        -------
        >>> BitString('1000').replace(1, BitString('101'))
        BitString('1101')
        >>> BitString('1111').replace(2, BitString('0'))
        BitString('1101')
        """
        if len(str(value)) > len(str(self)[offset:]):
            raise ValueError("Value too large for requested offset")
        bit_list = list(self.bitstring)
        to_insert = str(value)
        bit_list[offset : offset + len(to_insert)] = to_insert
        return BitString("".join(bit_list))

    @classmethod
    def concat(cls, *bitstrings: "BitString") -> "BitString":
        """
        Concatenate two or more BitStrings together as strings to produce a new
        BitString

        Parameters
        ----------
        *bitstrings : BitString
            BitString instances to concatenate

        Example
        -------
        >>> BitString.concat(BitString('10'), BitString('10'))
        BitString('1010')
        """
        if any(b for b in bitstrings if not isinstance(b, BitString)):
            raise TypeError("BitString.concat only operates on BitStrings")
        concatenated_string = "".join(map(str, bitstrings))
        return BitString(concatenated_string, fill=len(concatenated_string))

    def append(self, *others: "BitString") -> "BitString":
        """
        Join the string representation of other BitStrings onto this one.

        Note that unlike str.join, this method is variadic.

        Parameters
        ----------
        *others: BitString
            More BitStrings to append onto this instance

        Example
        -------
        >>> BitString('10').append(BitString('10'), BitString('01'))
        BitString('101001')
        """
        return BitString.concat(self, *others)

    def decode_bcd(self, *bounds: Tuple[int, int]) -> int:
        """
        Given bounds of each digit, slice the BitString to produce a place-value
        decimal number from those slices. First bound is most-significant bit.

        Parameters
        ----------
        * bounds: Tuple[int, int]
            The boundaries for each digit. Must be castable to a `slice`.

        Examples
        -------
        >>> BitString('10011001').decode_bcd((0, 4), (4, 8))
        99
        >>> BitString('10011001').decode_bcd((1, 4), (4, 8))
        19
        >>> BitString('10011001').decode_bcd((4, 8), (1, 4))
        91
        """
        if not bounds:
            raise ValueError("must pass at least one bound to decode BCD")
        digits = [self[slice(*b)] for b in bounds]
        return sum(10 ** i * d for i, d in enumerate(digits[::-1]))

    @classmethod
    def encode_bcd(
        cls, value: int, sizes: Collection[int] = (4, 4)
    ) -> "BitString":
        """
        Given a decimal value and a tuple of sizes (number of bits) for each
        digit, return a BitString with BCD encoding of that value. Defaults to
        8-bit encoding of 2 decimal digits in 4 bits each.

        Parameters
        ----------
        value: int
            The decimal value to encode
        sizes: Iterable[int, ...], default (4, 4)
            The number of bits to use to encode each digit

        Examples
        -------
        >>> BitString.encode_bcd(19)
        BitString('00011001')
        >>> BitString.encode_bcd(19, (3, 4))
        BitString('0011001')
        """
        if not isinstance(value, int) or value < 0:
            raise TypeError("Can only encode a positive (unsigned) int")
        n_digits = len(sizes)
        string_value = str(value).zfill(n_digits)
        if len(string_value) > n_digits:
            raise ValueError(
                f"Value {value} too large for number of digits implied by "
                f"sizes ({n_digits})."
            )
        return BitString.concat(
            *(
                BitString(int(digit), fill=bits, raise_on_overflow=True)
                for digit, bits in zip(string_value, sizes)
            )
        )

    def __repr__(self) -> str:
        return f"BitString('{self.bitstring}')"

    def __str__(self) -> str:
        return self.bitstring
